collection = {
    "https://neurohive.io/en/feed/",
    "http://feeds.feedburner.com/MachineLearningMastery",
    "https://www.fast.ai/index.xml",
    "http://ruder.io/rss/index.rss",
    "https://www.reddit.com/r/MachineLearning/.rss",
    "https://embracethered.com/blog/index.xml",
    "https://www.reddit.com/r/alife/.rss",
    "https://googleresearch.blogspot.com/atom.xml",
    "https://blog.einstein.ai/rss/",
    "https://www.louisbouchard.ai/rss/",
    "http://hunch.net/?feed=rss2",
}
