collection = {
    "http://roguelikedeveloper.blogspot.com/feeds/posts/default",
    "http://engineering.riotgames.com/rss.xml",
    "http://blog.ihobo.com/atom.xml",
    "http://www.raphkoster.com/feed/",
    "https://www.deconstructoroffun.com/blog?format=rss",
    "https://heredragonsabound.blogspot.com/feeds/posts/default",
    "https://godotengine.org/rss.xml",
    "http://www.filfre.net/feed/",
    "https://pvigier.github.io/rss.xml",
    "http://www.reddit.com/r/proceduralgeneration/.rss",
    "https://www.gridsagegames.com/blog",
    "http://www.gamedevblog.com/index.rdf",
    "http://fabiensanglard.net/rss.xml",
    "http://www.indiegames.com/blog/atom.xml",
    "http://www.zenofdesign.com/feed/",
    "https://lostgarden.home.blog/feed/",
    "http://www.reddit.com/r/gameai/.rss",
    "http://greyaliengames.com/blog/feed/",
    "http://www.reddit.com/r/roguelikedev/.rss",
    "https://heavens-vault-game.tumblr.com/rss",
    "https://tech.spaceapegames.com/feed/",
    "https://www.boristhebrave.com/feed/",
    "http://0fps.wordpress.com/feed/",
    "http://www.kloonigames.com/blog/feed/",
    "http://www.the-working-man.org/feeds/posts/default",
    "http://andrewfray.wordpress.com/feed/",
    "http://simblob.blogspot.com/feeds/posts/default",
    "https://www.reddit.com/r/alife/.rss",
    "http://www.alanzucconi.com/feed/",
}
